/*
 * ti113x.h 1.1 1996/11/20 07:49:24 (David Hinds)
 */

#ifndef _LINUX_TI113X_H
#define _LINUX_TI113X_H

/* Register definitions for TI 113X PCI-to-CardBus bridges */

#define TI113X_DCR			0x0092
#define  TI113X_DCR_IMODE_MASK		0x0006
#define  TI113X_DCR_IMODE_ISA		0x0002
#define  TI113X_DCR_IMODE_SERIRQ	0x0004

#endif /* _LINUX_TI113X_H */

